import typing

import numpy as np
import pandas as pd
import pingouin as pg
import sklearn.decomposition
import statsmodels.api as sm
import statsmodels.formula.api


def cronbachs_alpha(
    ratings: typing.Sequence,
    *,
    axis: int = 1,
) -> typing.Tuple[float, typing.Dict]:
    r"""Calculate Cronbach's alpha.

    The Cronbach coefficient quantifying interrater agreement.
    Returns alpha as a float
    and additional information specific to this measure
    collated into a dictionary.

    Cronbach's alpha generalizes Cohen's kappa
    and can handle three or more answers per variable.
    It is suitable for Likert type scale answers.
    A `blogpost on congeneric reliability`_ states
    that Cronbach’s alpha assumes essential tau-equivalence
    and underestimates reliability.
    A tau-equivalent measurement model
    is a special case of a congeneric measurement model
    with all loadings equal :cite:p:`cronbach-alpha-wikipedia`.

    A simplified formula is given in :cite:t:`Hilsdorf`
    that relates the measure to the average reliability:

    .. math::
       \alpha_{st} = \frac{N \times \bar{r}} {1 + (N - 1) \times \bar{r}}

    where

       - :math:`N` is the number of items (labelled chunks)
       - :math:`\bar{r}` is the average correlation between the items

    .. _blogpost on congeneric reliability: http://evaluationdashboard.com/index.php/2012/09/22/congeneric_reliability_r/

    Args:
        ratings: ratings.
            When given as a 1-dimensional array,
            it is treated as a row vector
        axis: axis along which the rater confidence is computed.
            A value of ``1``
            assumes stimuli as rows

    Returns:
         Cronbach's alpha and additional results lumped into dict

    """  # noqa: E501
    ratings = np.atleast_2d(np.array(ratings))
    n_items = ratings.shape[axis]  # K
    total_score = np.sum(ratings, axis=axis)  # X = ∑ Y_i
    variance_sum = np.var(ratings, axis=1 - axis, ddof=1).sum()  # ∑ var(Y_i)
    total_variance = total_score.var(ddof=1)  # var(X)
    alpha = n_items / (n_items - 1) * (1 - variance_sum / total_variance)

    result = {"total_variance": total_variance}
    return alpha, result


def congeneric_reliability(
    ratings: typing.Sequence,
    *,
    axis: int = 1,
) -> typing.Tuple[float, typing.Dict]:
    r"""Congeneric reliability coefficient.

    Extracts the first Principal Component as a measurement model
    and extracts the congeneric reliability coefficient.
    The implementation here is specific
    for a Principal Component Analysis with one factor.
    For implementation details see the wikipedia
    :cite:p:`congeneric-reliability-wikipedia`.

    Args:
        ratings: ratings.
            When given as a 1-dimensional array,
            it is treated as a row vector
        axis: axis along which the rater confidence is computed.
            A value of ``1``
            assumes stimuli as rows

    Returns:
        Congeneric Reliability and additional results lumped into dict

    """
    ratings = np.atleast_2d(np.array(ratings))
    if axis == 0:
        ratings = ratings.T
    pca = sklearn.decomposition.PCA(n_components=1)
    pca.fit(ratings)
    loadings = pca.components_.T * np.sqrt(pca.explained_variance_)  # λ_i
    loadings_sum = loadings.sum() ** 2  # (∑ λ_i)²
    total_score = np.sum(ratings, axis=1)  # X = ∑ Y_i
    total_variance = total_score.var(ddof=1)  # var(X)

    result = {}
    result["var. explained"] = pca.explained_variance_ratio_
    result["sing. values"] = pca.singular_values_

    reliability = loadings_sum / total_variance
    return reliability, result


def intra_class_correlation(
    ratings: typing.Sequence,
    *,
    axis: int = 1,
    icc_type: str = "ICC_1_1",
    anova_method: str = "pingouin",
) -> typing.Tuple[float, typing.Dict]:
    r"""Intraclass Correlation.

    Intraclass correlation calculates rating reliability by relating
    (i) variability of different ratings of the same subject to
    (ii) the total variation across all ratings and all items.

    The model is based on analysis of variance,
    and ratings must at least be ordinally scaled.

    CCC_ is conceptually and numerically related to the ICC.
    For an implementation see :func:`audmetric.concordance_cc`.

    Args:
        ratings: ratings.
            When given as a 1-dimensional array,
            it is treated as a row vector
        axis: axis along which the rater confidence is computed.
            A value of ``1``
            assumes stimuli as rows
        icc_type: ICC Method, see description below
        anova_method: method for ANOVA calculation,
            can be ``"pingouin"`` or ``"statsmodels"``

    .. _CCC: https://en.wikipedia.org/wiki/Concordance_correlation_coefficient

    Returns:
        icc and additional results lumped into dict

    Notes:
        :cite:ts:`Shrout/Fleiss:1979`
        describe a whole family of forms of coefficients
        that must be selected according
        to following criteria :cite:t:`Koo/Li:2016`:

        **Parametrization**

        :cite:t:`Koo/Li:2016` detail the parametrization:

        *Rater type: 1 versus k raters:*

        "This selection depends
        on how the measurement protocol
        will be conducted in actual application.
        For instance,
        if we plan to use the mean value of 3 raters as an assessment basis,
        the experimental design of the reliability study
        should involve 3 raters,
        and the 'mean of k raters' type should be selected.

        *Rating definition: absolute or consistency*

        "For both 2-way random- and 2-way mixed-effects models,
        there are 2 ICC definitions:
        “absolute agreement” and “consistency.”
        Selection of the ICC definition
        depends on whether we consider absolute agreement
        or consistency between raters to be more important.
        Absolute agreement concerns
        if different raters assign the same score
        to the same subject.
        Conversely,
        consistency definition concerns
        if raters’ scores to the same group of subjects
        are correlated in an additive manner.
        Consider an interrater reliability study of 2 raters as an example.
        In this case,
        consistency definition concerns the degree
        to which one rater’s score (y)
        can be equated to another rater’s score (x)
        plus a systematic error (c)
        (ie, :math:`y = x + c`),
        whereas absolute agreement
        concerns about the extent to which y equals x.

        *Model Choice*

        - One-Way Random-Effects Model
            In this model,
            each subject is rated by a different set of raters
            who were randomly chosen
            from a larger population of possible raters.
            This is used when different sets of raters rate different items
        - Two-Way Random-Effects Model
            If we randomly select our raters
            from a larger population of raters
            with similar characteristics
        - Two-Way Mixed-Effects Model
            We should use the 2-way mixed-effects model
            if the selected raters are the only raters of interest

        .. table::

            +----------------------+------------+--------------------+------------------------+
            | Parametrization      | Rater Type | Rating definition  | Model Choice           |
            +======================+============+====================+========================+
            | ICC(1,1) - icc\_1\_1 | single     | absolute agreement | One-way random effects |
            +----------------------+------------+--------------------+------------------------+
            | ICC(2,1) - icc\_2\_1 | single     | absolute agreement | Two-way random effects |
            +----------------------+------------+--------------------+------------------------+
            | ICC(3,1) - icc\_3\_1 | single     | consistency        | Two-way mixed effects  |
            +----------------------+------------+--------------------+------------------------+
            | ICC(1,k) - icc\_2\_k | k          | absolute agreement | One-way random effects |
            +----------------------+------------+--------------------+------------------------+
            | ICC(2,k) - icc\_2\_k | k          | absolute agreement | Two-way random effects |
            +----------------------+------------+--------------------+------------------------+
            | ICC(3,k) - icc\_3\_k | k          | consistency        | Two-way mixed effects  |
            +----------------------+------------+--------------------+------------------------+

        **Interpretation**

        Interpretation Conventions vary a between several sources,
        and depend also on the use case.
        We list values from the papers of :cite:ts:`Hallgren:2012`
        and :cite:ts:`Koo/Li:2016` respectively:

        .. table::

            +-----------+----------------+
            | Value     | Interpretation |
            +===========+================+
            | < 0.4     | poor           |
            +-----------+----------------+
            | .40 - .59 | fair           |
            +-----------+----------------+
            | .60 - .74 | good           |
            +-----------+----------------+
            | .75 - 1.0 | excellent      |
            +-----------+----------------+

        .. table::

            +-----------+----------------+
            | Value     | Interpretation |
            +===========+================+
            | < 0.5     | poor           |
            +-----------+----------------+
            | .50 - .75 | moderate       |
            +-----------+----------------+
            | .75 - .9  | good           |
            +-----------+----------------+
            | >.9 - 1.0 | excellent      |
            +-----------+----------------+

        **Formulas**:

        Shrout & Fleiss formula (1979, p. 423ff):

        .. math::
          \text{ICC}(1,1) = \frac{(\text{bms} - \text{wms})}{ (\text{bms} + (k - 1) * \text{wms})}

        .. math::
          \text{ICC}(2,1) = \frac{(\text{bms} - \text{ems})}{ (\text{bms} + (k - 1) * \text{ems} + k * (\text{jms} - \text{ems}) / n)}

        .. math::
          \text{ICC}(3,1) = \frac {(\text{bms} - \text{ems})}{ (\text{bms} + (k - 1) * \text{ems})}

        .. math::
          \text{ICC}(1,k) = \frac{(\text{bms} - \text{wms})}{\text{bms}}

        .. math::
          \text{ICC}(2,k) = \frac{(\text{bms} - \text{ems})}{ (\text{bms} + (\text{jms} - \text{ems}) / n)}

        .. math::
          \text{ICC}(3,k) = \frac{(\text{bms} - \text{ems})}{\text{bms}}

        where

        - :math:`\text{bms}` is the between items ("targets")
          mean sqeare of the underlying 2-factor Anova
        - :math:`\text{wms}` is the within items ("target") mean square
        - :math:`\text{jms}` is the between raters ("judges")
          mean square of the underlying 2-factor Anova
        - :math:`\text{ems}` is the error/residual mean square
        - :math:`k` is the number of raters
        - :math:`n` is the number of items

    """  # noqa: E501
    if not isinstance(ratings, pd.DataFrame):
        df = pd.DataFrame(np.atleast_2d(np.array(ratings)))
    else:
        df = ratings

    if axis == 0:
        df = df.T

    def _anova(df_long: pd.DataFrame, anova_method: str = "pingouin") -> pd.DataFrame:
        """Helper to get the anova table.

        Note that pingouin is currently default as statsmodels.
        ols from the statsmodels package is slow under many circumstances
        """
        anova_methods = ["statsmodels", "pingouin"]

        if anova_method not in anova_methods:
            raise ValueError(f"Anova method '{anova_method}' not in {anova_methods}.")

        if anova_method == "pingouin":
            # calculate and remap
            anova_table = pg.anova(
                data_long, dv="rating", between=["item", "rater"], ss_type=2
            )
            anova_table.columns = ["Source", "sum_sq", "df", "mean_sq", "np2"]
            return anova_table
            # aov_pg for wine:
            # Source         SS  DF         MS  np2
            # 0          item  188.21875   7  26.888393  1.0
            # 1         rater    7.34375   3   2.447917  1.0
            # 2  item * rater   47.90625  21   2.281250  1.0
            # 3      Residual    0.00000   0        NaN  NaN
        elif anova_method == "statsmodels":
            formula = "rating ~ C(item) * rater"
            model = statsmodels.formula.api.ols(formula=formula, data=data_long)
            result = model.fit()
            anova_table = sm.stats.anova_lm(result, typ=1)
            anova_table = sm.stats.anova_lm(result, typ=1)
            return anova_table
            # (Pdb) anova_table ols for wine
            #                  df        sum_sq    mean_sq    F  PR(>F)
            # C(item)         7.0  1.882188e+02  26.888393  0.0     NaN
            # rater           3.0  7.343750e+00   2.447917  0.0     NaN
            # C(item):rater  21.0  4.790625e+01   2.281250  0.0     NaN
            # Residual        0.0  9.138719e-27        inf  NaN     NaN
            # data must be (5152, 3)
        elif anova_method not in anova_methods:  # pragma: no cover
            raise ValueError("Undefined Anova method!")

    # data_long = df.reset_index().melt(id_vars=targets, value_name=ratings)
    # data_long = df.melt(ignore_index=False)
    data_long = df.melt(ignore_index=False, var_name="rater", value_name="rating")
    data_long["item"] = data_long.index

    # convert back to wide format to see how many raters dropped:
    data_wide = data_long.pivot_table(index="item", columns="rater", values="rating")

    # delete missing data in listwise manner
    nan_count = data_wide.isna().sum().sum()
    print(f"We have {nan_count} missing data points")
    print("Deleting them!")
    if nan_count > 0:
        data_wide = data_wide.dropna(axis=0, how="any")

    # back to long
    data_long = data_wide.melt(
        ignore_index=False, var_name="rater", value_name="rating"
    )
    data_long["item"] = data_long.index
    # drop missing
    data_long = data_long[data_long.columns].dropna()
    data_long = data_long.reset_index(drop=True)

    anova_table = _anova(data_long, anova_method=anova_method)

    bms = anova_table["mean_sq"].iloc[0]
    wms = (anova_table["sum_sq"].iloc[1] + anova_table["sum_sq"].iloc[2]) / (
        anova_table["df"].iloc[1] + anova_table["df"].iloc[2]
    )
    jms = anova_table["mean_sq"].iloc[1]
    ems = anova_table["mean_sq"].iloc[2]

    k = data_long["rater"].nunique()
    n = data_long["item"].nunique()

    # Calculate ICCs
    # Shrout & Fleiss Formula ICC(1,1), p. 423
    icc_1_1 = (bms - wms) / (bms + (k - 1) * wms)
    # Shrout & Fleiss Formula ICC(2,1), p. 423
    icc_2_1 = (bms - ems) / (bms + (k - 1) * ems + k * (jms - ems) / n)
    # Shrout & Fleiss Formula ICC(3,1), p. 423
    icc_3_1 = (bms - ems) / (bms + (k - 1) * ems)
    # Shrout & Fleiss Formula ICC(1,k), p. 42x
    icc_1_k = (bms - wms) / bms
    # Shrout & Fleiss Formula ICC(2,k), p. 42x
    icc_2_k = (bms - ems) / (bms + (jms - ems) / n)
    # Shrout & Fleiss Formula ICC(3,k), p. 42x
    icc_3_k = (bms - ems) / bms

    # Create output dataframe
    # stats = {
    icc_types = ["ICC_1_1", "ICC_2_1", "ICC_3_1", "ICC_1_k", "ICC_2_k", "ICC_3_k"]
    icc_rater_type = 3 * ["single"] + 3 * ["average"]
    icc_rating_type = ["absolute", "consistency", "consistency"] * 2
    icc_effect_type = ["-", "random", "fixed"] * 2
    icc = [icc_1_1, icc_2_1, icc_3_1, icc_1_k, icc_2_k, icc_3_k]

    results = pd.DataFrame(
        [icc_types, icc_rater_type, icc_rating_type, icc_effect_type, icc]
    ).T

    vars = ["icc_type", "rater type", "rating type", "anova effect type", "icc"]
    results_df = pd.DataFrame(
        [icc_types, icc_rater_type, icc_rating_type, icc_effect_type, icc], index=vars
    ).T

    results_table = results_df.to_dict("records")

    #     "Description": [
    #         "Single raters absolute",
    #         "Single random raters",
    #         "Single fixed raters",
    #         "Average raters absolute",
    #         "Average random raters",
    #         "Average fixed raters",
    #     ],
    #     "ICC": [icc_1_1, icc_2_1, icc_3_1, icc_1_k, icc_2_k, icc_3_k],
    # }

    icc_dict = [x for x in results_table if icc_type == x["icc_type"]]
    icc = icc_dict[0]["icc"]

    results = {
        "icc_dict": icc_dict,
        "results_table": results_table,
        "anova_table": anova_table,
    }
    return icc, results



import toml

import audeer


config = toml.load(audeer.path("..", "pyproject.toml"))

# Project -----------------------------------------------------------------
project = config["project"]["name"]
author = ", ".join(author["name"] for author in config["project"]["authors"])
copyright = "audEERING GmbH"
version = audeer.git_repo_version()
title = "Documentation"


# General -----------------------------------------------------------------
master_doc = "index"
extensions = []
source_suffix = ".rst"
exclude_patterns = [
    "api-src",
    "build",
    "Thumbs.db",
    ".DS_Store",
    "**.ipynb_checkpoints",
]
pygments_style = None
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",  # support for Google-style docstrings
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx_autodoc_typehints",
    "sphinx_copybutton",  # for "copy to clipboard" buttons
    "sphinxcontrib.bibtex",
    "sphinx_apipages",
]


bibtex_bibfiles = ["refs.bib"]
bibtex_default_style = "alpha"


# Ignore package dependencies during building the docs
autodoc_mock_imports = [
    "tqdm",
]
autodoc_type_aliases = {
    "ArrayLike": "ArrayLike",
}

# Reference with :ref:`data-header:Database`
autosectionlabel_prefix_document = True
autosectionlabel_maxdepth = 2

# Do not copy prompot output
copybutton_prompt_text = r">>> |\.\.\. "
copybutton_prompt_is_regexp = True

# Disable Gitlab as we need to sign in
linkcheck_ignore = [
    "https://gitlab.audeering.com",
    r".*evaluationdashboard.com/index.php/2012/09/22/*",
    "https://pypi.org/project/audpsychometric/",  # no package published yet
]


# HTML --------------------------------------------------------------------
html_theme = "sphinx_audeering_theme"
html_theme_options = {
    "display_version": True,
    "logo_only": False,
    "footer_links": False,
}
html_context = {
    "display_github": True,
}
html_title = title


# -- Intersphinx ------------------------------------------------
intersphinx_mapping = {
    "audmetric": ("https://audeering.github.io/audmetric/", None),
    "matplotlib": ("http://matplotlib.org/", None),
    "numpy": ("http://docs.scipy.org/doc/numpy/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "scipy": ("http://docs.scipy.org/doc/scipy/reference/", None),
    "seaborn": ("https://seaborn.pydata.org/", None),
    "sklearn": ("http://scikit-learn.org/stable", None),
    "python": ("https://docs.python.org/3/", None),
}



from audpsychometric.core.datasets import list_datasets
from audpsychometric.core.datasets import read_dataset
from audpsychometric.core.gold_standard import agreement_categorical
from audpsychometric.core.gold_standard import agreement_numerical
from audpsychometric.core.gold_standard import evaluator_weighted_estimator
from audpsychometric.core.gold_standard import mode
from audpsychometric.core.gold_standard import rater_agreement
from audpsychometric.core.reliability import congeneric_reliability
from audpsychometric.core.reliability import cronbachs_alpha
from audpsychometric.core.reliability import intra_class_correlation


# Disencourage from audpsychometric import *
__all__ = []

# Dynamically get the version of the installed module
try:
    import importlib.metadata

    __version__ = importlib.metadata.version(__name__)
except Exception:  # pragma: no cover
    importlib = None  # pragma: no cover
finally:
    del importlib



import numpy as np
import pytest

import audpsychometric


@pytest.fixture(scope="function")
def df_holzinger_swineford():
    df_dataset = audpsychometric.read_dataset("HolzingerSwineford1939")
    cols_use = [col for col in df_dataset.columns if col.startswith("x")]
    df = df_dataset[cols_use].astype(np.float32)
    return df


@pytest.fixture(scope="session", autouse=True)
def fixture_session():
    # always run this once when a session is started
    print("\nenter test session")
    yield
    # always run this once when a session has ended
    print("\nleave test session")


@pytest.fixture(scope="function")
def fixture_function():
    # run this before every test with this decorator:
    # @pytest.mark.usefixtures('fixture_function')
    print("\nenter test function")
    yield
    # run this after every test with this decorator:
    # @pytest.mark.usefixtures('fixture_function')
    print("\nleave test function")



import typing

import numpy as np
import pandas as pd

import audmetric


def agreement_categorical(
    ratings: typing.Sequence,
    *,
    axis: int = 1,
) -> typing.Union[float, np.ndarray]:
    r"""Confidence score for categorical ratings.

    The agreement for categorical data
    is given by the fraction of raters per item
    with the rating being equal to the mode
    given by :func:`audpsychometric.mode`.
    ``None`` and ``nan`` values are ignored per item.

    Args:
        ratings: ratings.
            When given as a 1-dimensional array,
            it is treated as a row vector
        axis: axis along which the agreement is computed.
            A value of ``1``
            assumes stimuli as rows
            and raters as columns

    Returns:
        categorical agreement score(s)

    Examples:
        >>> agreement_categorical([0, 1])
        0.5
        >>> agreement_categorical(["a", "b"])
        0.5
        >>> agreement_categorical([1, 1, np.nan])
        1.0

    """
    ratings = np.atleast_2d(np.array(ratings))

    def _agreement(x):
        x = _remove_empty(x)
        return np.sum(x == _mode(x)) / len(x)

    return _value_or_array(np.apply_along_axis(_agreement, axis, ratings))


def agreement_numerical(
    ratings: typing.Sequence,
    minimum: float,
    maximum: float,
    *,
    axis: int = 1,
) -> typing.Union[float, np.ndarray]:
    r"""Confidence score for numerical ratings.

    .. math::
        \text{agreement}(\text{ratings}) =
        \max(
        0, 1 - \frac{\text{std}(\text{ratings})}
        {\text{maximum} - \frac{1}{2} (\text{minimum} + \text{maximum})}
        )

    with :math:`\text{std}` the population standard deviation of the ratings.
    ``nan`` values are ignored per item.

    Args:
        ratings: ratings.
            When given as a 1-dimensional array,
            it is treated as a row vector
        minimum: lower limit of possible rating value
        maximum: upper limit of possible rating value
        axis: axis along which the agreement is computed.
            A value of ``1``
            assumes stimuli as rows
            and raters as columns

    Returns:
        numerical agreement score(s)

    Examples:
        >>> agreement_numerical([0, 1], 0, 1)
        0.0
        >>> agreement_numerical([0, 1], 0, 2)
        0.5
        >>> agreement_numerical([0, 0], 0, 1)
        1.0
        >>> agreement_numerical([0, np.nan], 0, 1)
        1.0

    """
    ratings = np.atleast_2d(np.array(ratings))

    def _agreement(row):
        cutoff_max = maximum - 1 / 2 * (minimum + maximum)
        return max([0.0, 1 - np.nanstd(row, ddof=0) / cutoff_max])

    return _value_or_array(np.apply_along_axis(_agreement, axis, ratings))


def evaluator_weighted_estimator(
    ratings: typing.Sequence,
    *,
    axis: int = 1,
) -> typing.Union[float, np.ndarray]:
    r"""Evaluator weighted estimator (EWE) of raters' votes.

    The EWE is described in
    :cite:`coutinho-etal-2016-assessing` as follows:

    The EWE average of the individual ratings considers
    that each evaluator is subject to an individual amount of disturbance
    during the evaluation,
    by introducing evaluator-dependent weights
    that correspond to the correlation
    between the listener’s responses
    and the average ratings of all evaluators.

    See also `audformat#102`_ for implementation details.

    .. _audformat#102: https://github.com/audeering/audformat/issues/102

    Args:
        ratings: ratings.
            Has to contain more than one rater
            and more than one stimuli
        axis: axis along which the EWE is computed.
            A value of ``1``
            assumes stimuli as rows
            and raters as columns

    Returns:
        EWE over raters

    Examples:
        >>> evaluator_weighted_estimator([[1, 1, 0], [2, 2, 1]])
        array([0.66666667, 1.66666667])

    """
    ratings = np.array(ratings)
    agreements = rater_agreement(ratings, axis=axis)
    # Ensure columns represents different raters
    if axis == 0:
        ratings = ratings.T
    return _value_or_array(np.inner(ratings, agreements) / np.sum(agreements))


def mode(
    ratings: typing.Sequence,
    *,
    axis: int = 1,
) -> typing.Any:
    r"""Mode of categorical ratings.

    ``None`` and ``nan`` values are ignored per item.
    If the values are numerical
    and there are several winning categories,
    the average over those is returned,
    and rounded to the next higher integer.
    If the values are not numerical,
    the first occurring value is returned.

    Args:
        ratings: ratings.
            When given as a 1-dimensional array,
            it is treated as a row vector
        axis: axis along which the mode is computed.
            A value of ``1``
            assumes stimuli as rows
            and raters as columns

    Returns:
        mode over raters

    Examples:
        >>> mode([0, 0, 1])
        0
        >>> mode(["a", "a", "b"])
        'a'
        >>> mode([0, 1])
        1
        >>> mode(["a", "b"])
        'a'
        >>> mode([0, 2])
        1
        >>> mode(["a", "c"])
        'a'
        >>> mode([None, None, "a"])
        'a'

    """
    ratings = np.atleast_2d(np.array(ratings))
    return _value_or_array(
        np.apply_along_axis(lambda x: _mode(x, remove_nan=True), axis, ratings)
    )


def rater_agreement(
    ratings: typing.Sequence,
    *,
    axis: int = 1,
) -> np.ndarray:
    """Calculate rater agreements.

    Calculate the agreement of a rater
    by the Pearson correlation of a rater
    with the mean score of all other raters.

    This should not be confused with the agreement value
    that relates to a rated stimulus,
    e.g. :func:`audspychometric.agreement_numerical`.

    Args:
        ratings: ratings.
            Has to contain more than one rater
            and more than one stimuli
        axis: axis along which the rater agreement is computed.
            A value of ``1``
            assumes stimuli as rows
            and raters as columns

    Returns:
        rater agreements

    Examples:
        >>> rater_agreement([[1, 1, 0], [2, 2, 1]])
        array([1., 1., 1.])
        >>> rater_agreement([[1, 1, 0], [2, 2, 1], [2, 2, 2]])
        array([0.94491118, 0.94491118, 0.8660254 ])

    """
    ratings = np.array(ratings)

    # Ensure columns represents different raters
    if axis == 0:
        ratings = ratings.T

    # Remove stimuli (rows),
    # which miss ratings for one rater or more
    ratings = ratings[:, ~np.isnan(ratings).any(axis=0)]

    # Calculate agreement as Pearson Correlation Coefficient
    # between the raters' ratings
    # and the average ratings of all other raters
    agreements = []
    for n in range(ratings.shape[1]):
        ratings_selected_rater = ratings[:, n]
        average_ratings_other_raters = np.delete(ratings, n, axis=1).mean(axis=1)
        agreements.append(
            audmetric.pearson_cc(ratings_selected_rater, average_ratings_other_raters)
        )
    return np.array(agreements)


def _value_or_array(values: np.ndarray) -> typing.Union[float, np.ndarray]:
    r"""Convert single valued arrays to value.

    Squeeze array,
    and convert to single value,
    if it contains only a single entry.

    Args:
        values: input array

    Returns:
        converted array / object

    """
    values = values.squeeze()
    if values.ndim == 0 or values.shape == (1,):
        values = values.item()
    return values


def _mode(ratings: np.ndarray, *, remove_nan=False) -> typing.Any:
    """Mode of categorical values.

    Args:
        ratings: 1-dimensional array
        remove_nan: if ``True`` remove ``None`` and ``nan`` from ``ratings``

    Returns:
        mode

    """
    if remove_nan:
        ratings = _remove_empty(ratings)
    values, counts = np.unique(ratings, return_counts=True)
    # Find indices with maximum count
    idx = np.flatnonzero(counts == np.max(counts))
    try:
        # Take average over values with same count
        # and round to next integer
        mode = int(np.floor(np.mean(values[idx]) + 0.5))
    except TypeError:
        # If we cannot take the mean,
        # take the first occurrence
        first_occurence = np.min([np.where(ratings == value) for value in values[idx]])
        mode = ratings[first_occurence]
    return mode


def _remove_empty(ratings: np.ndarray) -> np.ndarray:
    r"""Remove empty ratings.

    Args:
        ratings: 1-dimensional array

    Returns:
        ratings without ``None`` and ``nan`` entries

    """
    return np.array([rating for rating in ratings if not pd.isnull(rating)])



import numpy as np
import pandas as pd
import pytest

import audpsychometric


def test_icc():
    """Test icc basic result validity."""
    df_dataset = audpsychometric.read_dataset("wine")

    data_wide = df_dataset.pivot_table(index="Wine", columns="Judge", values="Scores")

    icc_sm, _ = audpsychometric.intra_class_correlation(
        data_wide, anova_method="statsmodels"
    )
    for ratings in [data_wide, data_wide.values]:
        icc_pingouin, _ = audpsychometric.intra_class_correlation(ratings)
        assert np.isclose(icc_pingouin, 0.727, atol=1e-3)
        assert np.isclose(icc_sm, icc_pingouin, atol=1e-10)
        icc_pingouin_t, _ = audpsychometric.intra_class_correlation(ratings.T, axis=0)
        assert np.isclose(icc_pingouin_t, icc_pingouin, atol=1e-10)


def test_cronbachs_alpha():
    """Test cronbach's alpha return values for three raters."""
    df_dataset = audpsychometric.read_dataset("hallgren-table3")
    df = df_dataset[["Dep_Rater1", "Dep_Rater2", "Dep_Rater3"]]
    for ratings in [df, df.values]:
        alpha, result = audpsychometric.cronbachs_alpha(ratings)
        assert isinstance(result, dict)
        assert np.isclose(alpha, 0.8516, atol=1e-4)
        alpha_t, _ = audpsychometric.cronbachs_alpha(ratings.T, axis=0)
        assert np.isclose(alpha_t, alpha, atol=1e-10)


def test_congeneric_reliability(df_holzinger_swineford):
    """Test congeneric reliability."""
    for ratings in [df_holzinger_swineford, df_holzinger_swineford.values]:
        coefficient, result = audpsychometric.congeneric_reliability(ratings)
    assert np.isclose(coefficient, 0.9365, atol=1e-4)
    assert np.isclose(result["var. explained"][0], 0.3713, atol=1e-4)
    for ratings in [df_holzinger_swineford, df_holzinger_swineford.values]:
        coefficient, result = audpsychometric.congeneric_reliability(ratings)
        assert np.isclose(coefficient, 0.9365, atol=1e-4)
        assert np.isclose(result["var. explained"][0], 0.3713, atol=1e-4)
        coefficient_t, _ = audpsychometric.congeneric_reliability(ratings.T, axis=0)
        assert np.isclose(coefficient_t, coefficient, atol=1e-10)


@pytest.mark.xfail(raises=ValueError)
def test_anova_helper():
    """Test that unknown anova parametrization raises exception."""
    audpsychometric.intra_class_correlation(pd.DataFrame(), anova_method="bbbb")


def test_icc_nanremoval():
    """Cover nan removal if statement."""
    df_dataset = audpsychometric.read_dataset("HolzingerSwineford1939")
    df_dataset = df_dataset[[x for x in df_dataset.columns if x.startswith("x")]]
    nan_mat = np.random.random(df_dataset.shape) < 0.1
    audpsychometric.intra_class_correlation(df_dataset.mask(nan_mat))



__all__ = ["read_dataset", "list_dataset"]

import os

import pandas as pd


data_directory = os.path.dirname(os.path.realpath(__file__))
dataset_path = os.path.join(data_directory, "datasets.csv")
data_sets = pd.read_csv(dataset_path, sep=",")


def read_dataset(data_set_name: str) -> pd.DataFrame:
    r"""Read dataset identified by name.

    retrieves a test dataset from within the package.

    Args:
        data_set_name(str): dataset name

    Returns:
        dataframe containing dataset

    Examples:
        >>> df = read_dataset("wine")
        >>> df.head()
           Wine Judge  Scores
        0     1     A       1
        1     2     A       1
        2     3     A       3
        3     4     A       6
        4     5     A       6

    """
    ds = data_sets.loc[data_sets["dataset"] == data_set_name]

    fname = ds["fname"].values[0]
    fpath = os.path.join(data_directory, fname)
    df = pd.read_csv(fpath, sep=",")
    return df


def list_datasets():
    r"""List tests datasets available in package.

    Returns:
        dataframe listing available datasets

    Examples:
        >>> list_datasets()
                                                     fname  ...                   description
        dataset                                             ...
        statology                            statology.csv  ...      icc sample from web page
        hallgren-table5              Hallgren-Table-05.csv  ...    icc table from publication
        hallgren-table3              Hallgren-Table-03.csv  ...  kappa table from publication
        HolzingerSwineford1939  HolzingerSwineford1939.csv  ...                        lavaan
        Shrout_Fleiss               Shrout_Fleiss_1979.csv  ...            Dataset from paper
        wine                                      wine.csv  ...                 online source
        <BLANKLINE>
        [6 rows x 4 columns]

    """  # noqa: E501
    df_data_sets = data_sets.set_index("dataset")
    return df_data_sets



"""Test Dataset Module."""
import pandas as pd
import pytest

import audpsychometric


def test_list_datasets():
    """First basic dataset is available in dataset list."""
    df_datasets = audpsychometric.list_datasets()
    assert "statology" in df_datasets.index


@pytest.mark.parametrize(
    "dataset",
    [
        "statology",
        "hallgren-table5",
        "hallgren-table3",
        "HolzingerSwineford1939",
        "Shrout_Fleiss",
        "wine",
    ],
)
def test_read_dataset(dataset):
    """Test functional requirement that a dataset can be read into dataframe."""
    df_dataset = audpsychometric.read_dataset(dataset)
    assert isinstance(df_dataset, pd.DataFrame)



import typing

import numpy as np
import pandas as pd
import pytest

import audpsychometric


def to_list_array_frame_series(
    x: list,
) -> typing.Tuple[list, np.ndarray, pd.DataFrame, typing.Optional[pd.Series]]:
    r"""Converts list to other input objects.

    It converts a list to an array,
    dataframe,
    and if is not a nested list (1-dimensional),
    to a series.
    For a 1-dimensional input,
    a dataframe with a single row is created.

    Args:
        x: values

    Returns:
        tuple containing values as list, array, dataframe,
        and maybe series

    """
    # list, array, dataframe
    outputs = [x, np.array(x), pd.DataFrame(np.atleast_2d(x))]
    # series
    if np.array(x).ndim == 1:
        outputs.append(pd.Series(x))
    return tuple(outputs)


@pytest.mark.parametrize(
    "ratings, axis, expected",
    [
        # axis = 0
        ([0], 0, 1.0),
        (["a"], 0, 1.0),
        ([0, 0], 0, np.array([1.0, 1.0])),
        ([[0, 0]], 0, np.array([1.0, 1.0])),
        ([[0], [0]], 0, 1.0),
        ([[None], ["a"]], 0, 1.0),
        # axis = 1
        ([0], 1, 1.0),
        (["a"], 1, 1.0),
        ([0, 0], 1, 1.0),
        ([[0, 0]], 1, 1.0),
        ([[0], [0]], 1, np.array([1.0, 1.0])),
        ([0, 0, 1], 1, 2 / 3),
        ([0, 0, 1, 1], 1, 0.5),
        (["a", "a", "b"], 1, 2 / 3),
        (["a", "a", "b", "b"], 1, 0.5),
        ([0, 1, 2], 1, 1 / 3),
        ([0, 1, 2, 2], 1, 0.5),
        (["a", "b", "c"], 1, 1 / 3),
        (["a", "b", "c", "c"], 1, 0.5),
        ([np.nan, 1, 1], 1, 1.0),
        ([np.nan, 0, 1], 1, 0.5),
        ([None, "a", "a"], 1, 1.0),
        ([None, "a", "b"], 1, 0.5),
        ([None, np.nan, 1], 1, 1.0),
    ],
)
def test_agreement_categorical(ratings, axis, expected):
    """Test agreement for categorical ratings.

    Args:
        ratings: ratings as list
        axis: axis along to compute agreement
        expected: expected agreement score(s)

    """
    for x in to_list_array_frame_series(ratings):
        np.testing.assert_equal(
            audpsychometric.agreement_categorical(x, axis=axis),
            expected,
            strict=True,
        )


# The expected agreement value for this test
# can be calculated by:
#
# def agreement(rating, minimum, maximum):
#     max_std = (maximum - minimum) / 2
#     std = np.std(rating)
#     std_norm = np.clip(std/max_std, 0, 1)
#     return 1 - std_norm
#
@pytest.mark.parametrize(
    "ratings, minimum, maximum, axis, expected",
    [
        # axis = 0
        ([0], 0, 1, 0, 1.0),
        ([0, 0], 0, 1, 0, np.array([1.0, 1.0])),
        ([[0, 0]], 0, 1, 0, np.array([1.0, 1.0])),
        ([[0], [0]], 0, 1, 0, 1.0),
        ([[0.3, 0.3, 0.3]], 0, 1, 0, np.array([1.0, 1.0, 1.0])),
        ([0, 1], 0, 1, 0, np.array([1.0, 1.0])),
        ([[0], [1]], 0, 1, 0, 0.0),
        ([[0], [1]], 0, 2, 0, 0.5),
        ([[1, 1]], 0, 1, 0, np.array([1.0, 1.0])),
        ([[1, 2, 3], [3, 4, 5]], 0, 10, 0, np.array([0.8, 0.8, 0.8])),
        ([[1, 2, 3], [3, 4, 4]], 0, 10, 0, np.array([0.8, 0.8, 0.9])),
        # axis = 1
        ([0], 0, 1, 1, 1.0),
        ([0, 0], 0, 1, 1, 1.0),
        ([[0, 0]], 0, 1, 1, 1.0),
        ([[0], [0]], 0, 1, 1, np.array([1.0, 1.0])),
        ([[1, 1]], 0, 1, 1, 1.0),
        ([[0.3, 0.3, 0.3]], 0, 1, 1, 1.0),
        ([0, 1], 0, 1, 1, 0.0),
        ([np.nan, 1], 0, 1, 1, 1.0),
        ([[0, 0, 0.1, 0.2]], 0, 1, 1, 0.83416876048223),
        ([[0, 0, 0.2, 0.4]], 0, 1, 1, 0.66833752096446),
        ([[0, 0, 0, 0, 0.2, 0.2, 0.4, 0.4]], 0, 1, 1, 0.66833752096446),
        ([[0, 0.4, 0.6, 1]], 0, 1, 1, 0.2788897449072021),
        ([[0, 0.33, 0.67, 1]], 0, 1, 1, 0.2531399060064863),
        ([[0, 1]], 0, 1, 1, 0.0),
        ([[0, 0, 1, 1]], 0, 1, 1, 0.0),
        (
            [[1, 2, 3], [3, 4, 5]],
            0,
            10,
            1,
            np.array([0.8367006838144548, 0.8367006838144548]),
        ),
        (
            [[1, 2, 3], [3, 4, 4]],
            0,
            10,
            1,
            np.array([0.8367006838144548, 0.9057190958417937]),
        ),
    ],
)
def test_agreement_numerical(ratings, minimum, maximum, axis, expected):
    """Test agreement for numerical ratings.

    If only a vector is given for ``ratings``,
    it should be treated as column vector.
    An value of 0 for ``axis``
    should compute the agreement scores along rows.

    Args:
        ratings: ratings as list
        minimum: lower limit of ratings
        maximum: upper limit of ratings
        axis: axis along to compute agreement
        expected: expected agreement score(s)

    """
    for x in to_list_array_frame_series(ratings):
        np.testing.assert_equal(
            audpsychometric.agreement_numerical(x, minimum, maximum, axis=axis),
            expected,
            strict=True,
        )


def test_rater_agreement(df_holzinger_swineford):
    """Test rater agreement."""
    # there is a very unrealible rater in this set with .24
    expected = np.array(
        [
            0.52203673,
            0.27524307,
            0.37017212,
            0.58070663,
            0.52538537,
            0.59513902,
            0.24573167,
            0.36905549,
            0.49478097,
        ],
    )
    np.testing.assert_allclose(
        audpsychometric.rater_agreement(df_holzinger_swineford),
        expected,
        strict=True,
    )


@pytest.mark.parametrize(
    "ratings, axis, expected",
    [
        # axis = 0
        ([0], 0, 0),
        ([0, 0], 0, np.array([0, 0])),
        ([[0, 0]], 0, np.array([0, 0])),
        ([[0], [0]], 0, 0),
        (["a"], 0, "a"),
        (["a", "a"], 0, np.array(["a", "a"])),
        ([["a", "a"]], 0, np.array(["a", "a"])),
        ([["a"], ["a"]], 0, "a"),
        ([[np.nan, 1], [2, np.nan]], 0, np.array([2, 1])),
        # axis = 1
        ([0], 1, 0),
        ([0, 0], 1, 0),
        ([0, 1], 1, 1),
        ([0, 0, 1], 1, 0),
        ([0, 1, 2], 1, 1),
        ([0, 1, 1], 1, 1),
        ([0, 2, 2, 1], 1, 2),
        ([0, 2, 2, 1, 1], 1, 2),
        ([0, 2, 2, 1, 1, 1], 1, 1),
        ([0, 2], 1, 1),
        ([[0, 0]], 1, 0),
        ([[0], [0]], 1, np.array([0, 0])),
        ([[1, 1]], 1, 1),
        (["a"], 1, "a"),
        (["a", "a"], 1, "a"),
        (["a", "b"], 1, "a"),
        (["b", "a"], 1, "b"),
        (["a", "a", "b"], 1, "a"),
        (["b", "a", "a"], 1, "a"),
        (["a", "b", "c"], 1, "a"),
        (["a", "c", "c", "b"], 1, "c"),
        (["a", "c", "c", "b", "b"], 1, "c"),
        (["a", "c", "c", "b", "b", "b"], 1, "b"),
        ([["a", "a"]], 1, "a"),
        ([["a"], ["a"]], 1, np.array(["a", "a"])),
        ([np.nan, np.nan, 1], 1, 1),
        ([np.nan, np.nan, None, None, 1], 1, 1),
        ([[np.nan, 1], [2, np.nan]], 1, np.array([1, 2])),
    ],
)
def test_mode(ratings, axis, expected):
    """Test mode over ratings.

    If the categories are integers,
    and there is no winner,
    mode should return the average of all winners.
    For other categories,
    the first item should be returned
    of the winning categories.

    Args:
        ratings: ratings as list
        axis: axis along to compute mode
        expected: expected mode values

    """
    for x in to_list_array_frame_series(ratings):
        np.testing.assert_equal(
            audpsychometric.mode(x, axis=axis),
            expected,
            strict=True,
        )


@pytest.mark.parametrize("axis", [0, 1])
def test_evaluator_weighted_estimator(df_holzinger_swineford, axis):
    """Test EWE over ratings.

    Args:
        df_holzinger_swineford: df_holzinger_swineford fixture
        axis: axis along to compute EWE

    """
    ewe = audpsychometric.evaluator_weighted_estimator(
        df_holzinger_swineford,
        axis=axis,
    )
    if axis == 0:
        expected = df_holzinger_swineford.values.shape[1]
    elif axis == 1:
        expected = df_holzinger_swineford.values.shape[0]
    assert ewe.shape[0] == expected
